const std = @import("std");
const vk = @import("vulkan");
const zm = @import("zmath");
const ai = @import("assimp.zig").c;

const Mesh = @import("Mesh.zig");
const Context = @import("Context.zig");
const Device = Context.Device;
const Instance = @import("Context.zig").Instance;
const Vertex = @import("utilities.zig").Vertex;
const StringUtils = @import("string_utils.zig");
const Texture = @import("Texture.zig");

const Self = @This();

allocator: std.mem.Allocator,

mesh_list: std.ArrayList(Mesh),
textures: std.ArrayList(Texture),
model: zm.Mat,

sampler_descriptor_sets: std.ArrayList(vk.DescriptorSet),

pub fn new(
    allocator: std.mem.Allocator,
    ctx: Context,
    graphics_command_pool: vk.CommandPool,
    texture_sampler: vk.Sampler,
    model_file: []const u8,
) Self {
    var new_mesh_model: Self = undefined;

    new_mesh_model.allocator = allocator;
    new_mesh_model.model = zm.identity();

    new_mesh_model.sampler_descriptor_sets = try std.ArrayList(vk.DescriptorSet)
        .init(allocator);

    const path = try StringUtils.concat("assets/models/", model_file, allocator);
    defer allocator.free(path);

    // Import model scene
    const scene = ai.aiImportFile(
        path.ptr,
        ai.aiProcess_Triangulate | ai.aiProcess_FlipUVs | ai.aiProcess_JoinIdenticalVertices,
    );
    defer ai.aiReleaseImport(scene);

    // Get array of all materials with 1:1 ID placement
    const texture_names = try ai.loadMaterials(allocator, scene);
    defer {
        for (0..texture_names.items.len) |i| {
            if (texture_names.items[i]) |texture_name| {
                allocator.free(texture_name);
            }
        }
        texture_names.deinit();
    }

    // Conversion from the material list IDs to our descriptor array IDs
    new_mesh_model.textures = try std.ArrayList(Texture).initCapacity(allocator, texture_names.items.len);

    // Loop over texture names and create textures for them
    for (texture_names.items) |texture_name| {
        if (texture_name != null) {
            // Create texture and set value to index of new texture
            new_mesh_model.textures.appendAssumeCapacity(try Texture.create(texture_name.?));
        } else {
            // If material had no texture, set to 0 to indicate no texture. Texture 0 will be reserver for a default texture
            // TODO Put the default texture somewhere else where it's shared
            new_mesh_model.textures.appendAssumeCapacity(try Texture.create("giraffe.png"));
        }
    }

    // Load in all our meshes
    new_mesh_model.mesh_list = try loadNode(
        allocator,
        ctx.instance,
        ctx.physical_device,
        ctx.device,
        ctx.graphics_queue.handle,
        graphics_command_pool,
        scene.*.mRootNode,
        scene,
    );

    return new_mesh_model;
}

pub fn destroy(self: *Self) void {
    for (0..self.mesh_list.items.len) |i| {
        self.mesh_list.items[i].destroy();
    }
    self.mesh_list.deinit();
}

pub fn getMesh(self: Self, idx: usize) !Mesh {
    if (idx >= self.mesh_list.len - 1) {
        return error.MeshIndexOutOfBounds;
    }

    return self.mesh_list.items[idx];
}

pub fn loadNode(
    allocator: std.mem.Allocator,
    instance: Instance,
    pdev: vk.PhysicalDevice,
    device: Device,
    transfer_queue: vk.Queue,
    transfer_command_pool: vk.CommandPool,
    node: *const ai.aiNode,
    scene: *const ai.aiScene,
) !std.ArrayList(Mesh) {
    var mesh_list = std.ArrayList(Mesh).init(allocator);

    // Go through each mesh at this node and create it, then add it to our mesh list
    for (0..node.mNumMeshes) |i| {
        // Load mesh here
        try mesh_list.append(try loadMesh(
            allocator,
            instance,
            pdev,
            device,
            transfer_queue,
            transfer_command_pool,
            scene.mMeshes[node.mMeshes[i]],
        ));
    }

    // Go through each node attached to this node and load it, then append their meshes to this node's mesh list
    for (0..node.mNumChildren) |i| {
        var new_list = try loadNode(
            allocator,
            instance,
            pdev,
            device,
            transfer_queue,
            transfer_command_pool,
            node.mChildren[i],
            scene,
        );
        defer new_list.deinit();

        try mesh_list.appendSlice(new_list.items[0..]);
    }

    return mesh_list;
}

pub fn loadMesh(
    allocator: std.mem.Allocator,
    instance: Instance,
    pdev: vk.PhysicalDevice,
    device: Device,
    transfer_queue: vk.Queue,
    transfer_command_pool: vk.CommandPool,
    mesh: *const ai.aiMesh,
) !Mesh {
    var vertices = try std.ArrayList(Vertex).initCapacity(allocator, mesh.mNumVertices);
    var indices = std.ArrayList(u32).init(allocator);
    defer vertices.deinit();
    defer indices.deinit();

    // Go through each vertex and copy it across to our vertices
    for (0..mesh.mNumVertices) |i| {
        var vertex: Vertex = undefined;

        // Set position
        vertex.pos = .{ mesh.mVertices[i].x, mesh.mVertices[i].y, mesh.mVertices[i].z };

        // Set texture coords (if they exist)
        if (mesh.mTextureCoords[0] != null) {
            vertex.tex = .{ mesh.mTextureCoords[0][i].x, mesh.mTextureCoords[0][i].y };
        } else {
            vertex.tex = .{ 0.0, 0.0 };
        }

        // Set colour (just use white for now)
        vertex.col = .{ 1.0, 1.0, 1.0 };

        // vertices.items[i] = vertex;
        vertices.appendAssumeCapacity(vertex);
    }

    // Iterate over indices through faces and copy across
    for (0..mesh.mNumFaces) |i| {
        // Get a face
        const face = mesh.mFaces[i];

        // Go through face's indices and add to list
        for (0..face.mNumIndices) |j| {
            try indices.append(face.mIndices[j]);
        }
    }

    return try Mesh.create(
        allocator,
        instance,
        pdev,
        device,
        transfer_queue,
        transfer_command_pool,
        vertices.items,
        indices.items,
        mesh.mMaterialIndex,
    );
}
